"""
Worker
======

A thread-based worker for crawling web pages. It pulls URLs to parse from
a queue, and places finished results into a list accessible by multiple
workers.
"""


from threading import Thread, Lock

from .queues import EmptyQueueError, MaxResultsExceededError
from .loggers import Logger
from .loaders import Loader


class WorkerStates():
    WAITING = "WAITING"
    STOPPED = "STOPPED"
    RUNNING = "RUNNING"
    READY = "READY"
    STATES = [WAITING, STOPPED, RUNNING, READY]


class Worker(WorkerStates, Thread):
    """
    :param queue: a Soup Stars queue, which only allows putting an item once
    :type queue: class:`soupstars.queue.Queue`
    :param results: a results queue
    :type results: class:`list`
    """

    states = WorkerStates

    @classmethod
    def start_all(kls, num, *args, **kwargs):
        """
        Create and start `num` workers. Arguments are passed to the worker's
        __init__ function.
        """

        workers = [kls(id=id, *args, **kwargs) for id in range(1, num+1)]
        for worker in workers:
            worker.start()
        return workers

    @classmethod
    def stop_all(kls, workers):
        for worker in workers:
            worker.stop()
        for worker in workers:
            worker.join()
        return workers

    @classmethod
    def all_waiting(kls, workers):
        return all([w.state == w.WAITING for w in workers])

    def __str__(self):
        return f"<{self.__class__.__name__} {self.id}>"

    def start(self):
        """
        Starts the worker, running until `stop` is called.
        """
        self.state = self.set_state(self.RUNNING)
        self.logger.okay(f"Starting {self}")
        super().start()

    # I think we need to get a lock on `self` while this is happening. Ie what
    # if it is being executed at the same time by two different calls?
    def set_state(self, state):
        if state not in self.STATES:
            raise AttributeError(f"Can not change state to {state}, must be \
                                   one of {self.STATES}")
        self.lock.acquire()
        if self.state == self.STOPPED:
            pass
        else:
            self.state = state
        self.lock.release()
        return self.state

    def stop(self):
        """
        Stops the worker from running after it has finished any work
        currently in process.
        """

        self.logger.okay("Got signal to stop")
        self.set_state(self.STOPPED)


class LinksWorker(Worker):
    """
    Processes a `links` queue
    """

    def __init__(self, id, links_queue, results_queue, parser):
        self.lock = Lock()
        self.id = id
        self.links_queue = links_queue
        self.results_queue = results_queue
        self.parser = parser
        self.state = self.READY
        self.logger = Logger(self)
        self.loader = Loader()
        super().__init__()

    def run(self):
        self.links_queue.put_all(self.parser.seeds(), ignore_duplicates=True)
        while not self.state == self.STOPPED:
            if self.results_queue.full():
                self.set_state(self.WAITING)
                continue
            try:
                url = self.links_queue.get(timeout=0.1)
                self.set_state(self.RUNNING)
            except EmptyQueueError:
                self.set_state(self.WAITING)
                continue

            try:
                response = self.loader.load(url)
                data, links = self.parser.parse(response)
                if data:
                    try:
                        self.results_queue.put(data)
                    except MaxResultsExceededError:
                        continue
                self.links_queue.put_all(links, ignore_duplicates=True)
                self.logger.okay(f"Processed {url}")
            except Exception as err:
                self.logger.fail(f"Failed processing {url}: {err}")
            finally:
                self.links_queue.task_done()
        else:
            self.logger.okay(f"Stopped")


class ResultsWorker(Worker):
    """
    Processes the `results` queue
    """

    def __init__(self, id, writer, results_queue):
        self.lock = Lock()
        self.id = id
        self.state = self.READY
        self.logger = Logger(self)
        self.writer = writer
        self.results_queue = results_queue
        super().__init__()

    def run(self):
        while not self.state == self.STOPPED:
            try:
                result = self.results_queue.get(timeout=0.1)
                self.set_state(self.RUNNING)
            except EmptyQueueError:
                self.set_state(self.WAITING)
                continue
            try:
                self.writer.write(result)
                self.logger.okay(f"Processed {result['url']}")
            except Exception as err:
                self.logger.fail(f"Failed writing {result['url']}: {err}")
            finally:
                self.results_queue.task_done()
        else:
            self.logger.okay(f"Stopped")



import os
import json


class Config(object):
    token = None
    host = "https://soupstars-cloud.herokuapp.com"
    home = os.path.join(os.path.expanduser('~'), '.soupstars')
    fname = 'config.json'


    def __init__(self, host=None, token=None, home=None, fname=None):
        self.name =  fname or os.environ.get('SOUPSTARS_FNAME') or self.fname
        self.home =  home or os.environ.get('SOUPSTARS_HOME') or self.home
        self.path = os.path.join(self.home, self.name)
        self.logs = os.path.join(self.home, 'logs')

        if not os.path.exists(self.home):
            os.mkdir(self.home)
        if not os.path.exists(self.logs):
            os.mkdir(self.logs)

        config = self.load()
        self.host = host or os.environ.get('SOUPSTARS_HOST') or config['host']
        self.token = token or os.environ.get('SOUPSTARS_TOKEN') or config['token']  # noqa

    def to_dict(self):
        return {
            "host": self.host,
            "token": self.token
        }

    def save(self):
        with open(self.path, 'w') as o:
            json.dump(self.to_dict(), o, indent=2)

    def load(self):
        if not os.path.exists(self.path):
            self.save()
        with open(self.path) as o:
            return json.load(o)



"""
Extract article links and metadata from nytimes.com
"""

from soupstars import data, follow, exclude


url = "https://www.nytimes.com"


@follow
def follow(url):
    return (url.domain == "www.nytimes.com") and (url.match("\d{4}\/\d{2}\/\d{2}"))


@exclude
def exclude(soup):
    return not soup.url.match("\d{4}\/\d{2}\/\d{2}")


@data
def title(soup):
    return soup.title()



"""
Client
======

Wrapper for the SoupStars web API.

>>> from soupstars import Client
>>> client = Client()
>>> resp = client.status()
>>> resp.json['healthy']
True

"""

import os
import requests
import json

from .config import Config


class Routes(object):
    health = "/health/"
    login = "/token/"
    register = '/register/'
    ls = '/list/'
    profile = '/profile/'
    pull = '/pull/'
    push = '/push/'
    run = '/run/'
    results = '/results/'
    results_create = '/results/create/'


class Client(object):
    routes = Routes()

    def __init__(self, token=None, **kwargs):
        self.config = Config(token=token, **kwargs)
        self.token = self.config.token

    def send(self, path, method='POST', data=None, files=None):
        url = self.config.host + path
        headers = {'Authorization': self.token}
        resp = requests.request(
            url=url,
            method=method,
            json=data,
            headers=headers,
            files=files
        )
        return resp

    def register(self, email, password):
        data = {
            'email': email,
            'password': password
        }
        return self.send(self.routes.register, data=data)

    def login(self, email, password):
        data = {
            'email': email,
            'password': password
        }
        return self.send(self.routes.login, data=data)

    def health(self):
        return self.send(self.routes.health, method='GET')

    def ls(self):
        return self.send(self.routes.ls)

    def profile(self):
        return self.send(self.routes.profile)

    def push(self, module):
        file_data = open(module, 'rb').read()
        files = {module: file_data}
        return self.send(self.routes.push, files=files)

    def pull(self, module):
        data = {'name': module}
        return self.send(self.routes.pull, data=data)

    def run(self, module, _async=False):
        data = {'name': module, 'async': _async}
        return self.send(self.routes.run, data=data)

    def results(self, module):
        data = {'name': module}
        return self.send(self.routes.results, data=data)

    def create_result(self, run_id, parser_id, data, status_code, url, errors):
        return self.send(self.routes.results_create, data={
            'result': {
                'run_id': run_id,
                'parser_id': parser_id,
                'data': data,
                'status_code': status_code,
                'url': url,
                'errors': errors
            }
        })



"""
Runner
======

The runner is responsible for executing the parser module built by a user.

It can be configured in a number of ways:
- select the number of threaded workers, so that multiple requests can be made concurrently and the results of those requests processed in parallel.
- only process a select number of pages.
- write results to various locations and formats
"""

import importlib
import time
from datetime import datetime as dt
from types import ModuleType

from .queues import LinksQueue, ResultsQueue
from .workers import LinksWorker, ResultsWorker
from .loggers import Logger
from .parsers import Parser
from .writers import JsonWriter, Writer


class Runner(object):

    def __init__(self, parser, writer='csv', max_results=0):
        self.writer = Writer.from_string(writer)
        self.parser = Parser.from_string(parser)
        self.results = ResultsQueue(max_results=max_results)
        self.queue = LinksQueue()
        self.logger = Logger(self)
        self.link_workers = []
        self.result_workers = []

    def __str__(self):
        return f"<Runner {self.parser.module.__name__}>"

    def finished(self):
        return ResultsWorker.all_waiting(self.result_workers) and \
               LinksWorker.all_waiting(self.link_workers)

    # TODO: pull up configuration (num workers)
    def run(self):
        """
        Starts running the parser. Creates the workers and waits for them to
        finish.
        """

        self.logger.okay(f"Starting run")
        self.writer.url = f"{self.parser.module.__name__}_run_{dt.isoformat(dt.now())}"
        self.link_workers = LinksWorker.start_all(4, links_queue=self.queue, results_queue=self.results, parser=self.parser)
        self.result_workers = ResultsWorker.start_all(1, writer=self.writer, results_queue=self.results)

        while not self.finished():
            continue
        else:
            LinksWorker.stop_all(self.link_workers)
            ResultsWorker.stop_all(self.result_workers)
            self.logger.okay("Stopped all workers")
            self.logger.okay("Finished run")



import click
import os
import sys
import textwrap
import getpass

from ..clients import Client
from .printers import jsonify, pythonify


@click.group()
@click.option('--token', '-t', help="Token to use. Default None")
@click.option('--host', '-h', default=None, help="Host to use")
@click.pass_context
def cloud(context, token, host):
    """
    Commands to interact with SoupStars cloud.
    """

    context.obj = Client(token=token, host=host)


@cloud.command()
@click.pass_obj
def health(client):
    """
    Print the status of the SoupStars api
    """

    resp = client.health()
    jsonify(resp.json())


@cloud.command()
@click.pass_obj
def config(client):
    """
    Print client configuration details
    """

    jsonify(client.config.to_dict())



@cloud.command()
@click.pass_obj
def login(client):
    """
    Log in with an existing email
    """

    email = input('Email: ')
    password = getpass.getpass(prompt='Password: ')
    resp = client.login(email=email, password=password)

    if resp.ok:
        client.config.token = resp.json()['token']
        client.config.save()

    jsonify(resp.json())


@cloud.command()
@click.pass_obj
def ls(client):
    """
    Show the parsers uploaded to SoupStars cloud
    """

    resp = client.ls()
    jsonify(resp.json())


@cloud.command()
@click.pass_obj
def register(client):
    """
    Register a new account on SoupStars cloud
    """

    email = input('Email: ')
    password = getpass.getpass(prompt='Password: ')
    password2 = getpass.getpass(prompt='Confirm password: ')

    if password != password2:
        print("Passwords did not match.")
        return

    resp = client.register(email=email, password=password)

    if resp.ok:
        client.config.token = resp.json()['token']
        client.config.save()

    jsonify(resp.json())


@cloud.command()
@click.pass_obj
def whoami(client):
    """
    Print the email address of the current user
    """

    resp = client.profile()
    jsonify(resp.json())


@cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to push")
@click.pass_obj
def push(client, module):
    """
    Push a parser to SoupStars cloud
    """

    resp = client.push(module)
    jsonify(resp.json())


@cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to pull")
@click.pass_obj
def pull(client, module):
    """
    Pull a parser from SoupStars cloud into a local module
    """

    resp = client.pull(module)
    data = resp.json()
    with open(data['parser']['name'], 'w') as o:
        o.write(data['module'])
    jsonify({"state": "done", "response": data})


@cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to create")
@click.pass_obj
def run(client, module):
    """
    Run a parser on SoupStars cloud
    """

    resp = client.run(module)
    jsonify(resp.json())


@cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to show")
@click.option('--json/--no-json', default=False, help="Show parser details in JSON")
@click.pass_obj
def show(client, module, json):
    """
    Show the contents of a parser on SoupStars cloud
    """

    resp = client.pull(module)

    if json:
        jsonify(resp.json())
    else:
        pythonify(resp.json()['module'])


@cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to test")
@click.pass_obj
def results(client, module):
    """
    Print results of a parser
    """

    resp = client.results(module)
    jsonify(resp.json())



import click
import os
import sys

from ..api import fetch, run as _run
from ..clients import Client
from ..loaders import Loader
from ..parsers import Parser, BeautifulSoupStar, Url
from ..version import __version__
from .printers import jsonify


@click.group()
def base():
    """
    CLI for managing Soup Stars parsers
    """

    if '.' not in sys.path:
        sys.path.append('.')


@base.command()
def system():
    """
    Print system configuration details
    """

    data = {
        "soupstars": {
            "version": __version__,
        },
        "python": {
            "version": f"{sys.version})".replace('\n', ''),
            "path": sys.executable,
        },
        "platform": sys.platform
    }

    jsonify(data)


@base.command()
@click.argument('parser')
@click.option('--url', '-u', required=True, help="Url to parse")
def debug(parser, url):
    """
    Open a python prompt with a parser result
    """

    parser = Parser.from_string(parser)
    soup = fetch(url)
    response = soup.response
    request = soup.request
    data, links = parser.parse(response)

    try:
        import IPython
        IPython.embed(colors="neutral")
    except ImportError:
        import code
        code.InteractiveConsole(locals=vars()).interact()


@base.command()
@click.argument('name')
@click.option('--template', '-t', required=False, default='nytimes', help="Name of the template to use. Default is `nytimes`")
@click.option('--force/--no-force', default=False, help="Overwrite an existing file")
def create(name, template, force):
    """
    Create a new parser from a template
    """

    parser = Parser.from_string(f'soupstars.examples.{template}')
    text = open(parser.module.__file__, 'r').read()
    if os.path.exists(name) and not force:
        raise FileExistsError(
            f"{name} already exists. Use --force to overwrite."
        )
    with open(name, 'w') as o:
        o.write(text)
        jsonify({"state": "done", "parser": name})


@base.command()
@click.argument('parser')
@click.option('--limit', '-l', required=False, default=100, help="Number of results to parse")
def run(parser, limit):
    """
    Run a parser
    """

    _run(parser, max_results=limit)
